/*
 * Copyright 2000 by Hans Reiser, licensing governed by reiserfs/README
 */

#define __USE_GNU
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <asm/types.h>
#include <sys/vfs.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <asm/types.h>

//#include "inode.h"
#include "io.h"
#include "misc.h"
#include "reiserfs_lib.h"


typedef __u32 blocknr_t;

#define MAX_OFFSET_V1 0xffffffff
#define MAX_OFFSET_V2 0xffffffffffffffff

/* searches.c */
//#define KEY_FOUND 1
//#define KEY_NOT_FOUND 0

//#define DIRECTORY_NOT_FOUND -1

//#define FILE_NOT_FOUND -1

#define UNACCESSED_MASK  0xff00
#define ACCESSED_MASK    0x00ff


/* main.c */
int main (int argc, char * argv []);


//
// options
//
extern int opt_verbose;
extern int opt_fsck;


/* fsck has modes */
#define FSCK_CHECK 1
#define FSCK_REBUILD 2
#define FSCK_FAST_REBUILD 3
#define FSCK_MAKE_SB 4
#define FSCK_SEARCH_KEY 5

#define REPLAY_JOURNAL 1

#define PASS0   0
#define PASS1   1
#define PASS1a  2
#define PASS2   2
extern int fsck_stage;

//
// replay journal modes
//
#define REPLAY_DEFAULT 0
#define REPLAY_ALL 1
#define NO_REPLAY 2
#define REPLAY_DEFAULT 0
#define REPLAY_ALL 1
#define NO_REPLAY 2
extern int opt_journal_replay;

extern struct key key_to_find;
extern __u64 offset_to_find;

#define STOP_DEFAULT 0
#define STOP_AFTER_PASS1 1
#define STOP_AFTER_PASS2 2
#define STOP_AFTER_SEMANTIC 3
#define STOP_AFTER_REPLAY 4
extern int opt_stop_point;

#define SCAN_USED_PART 0
#define SCAN_WHOLE_PARTITION 1
extern int opt_what_to_scan;

#define NO_LOST_FOUND 0
#define DO_LOST_FOUND 1
extern int opt_lost_found;

extern struct super_block g_sb;
extern struct reiserfs_transaction_handle g_th;

extern struct reiserfs_super_block * g_old_rs;
extern unsigned long SB_CURR_SIZE;

extern char ** g_disk_bitmap;
extern char ** g_new_bitmap;
extern char ** g_uninsertable_leaf_bitmap;
extern char ** g_formatted;
extern char ** g_unformatted;
extern char ** control_bitmap;
extern __u32 g_blocks_to_read_in_pass0;
extern __u32 g_blocks_to_read_in_pass1;

extern struct key root_key;

/* pass1.c */
/* handle indirect items modes*/
#define INSERT            1
#define POSTPONE          2
#define SAVE              3
#define INSERT_SEPARATELY 4

void build_the_tree (void);
extern int g_unaccessed_items;
int is_item_reachable (struct item_head * ih);
void mark_item_reachable (struct item_head * ih, struct buffer_head * bh);
void mark_item_unreachable (struct item_head * ih);

/*
extern inline int dir_entry_length (struct reiserfs_de_head * deh, struct item_head * ih, int pos_in_item)
{
    if (pos_in_item)
	return (deh_location (deh - 1) - deh_location (deh));
    return (ih_item_len (ih) - deh_location (deh));
}
*/

/* file.c */
struct si {
    struct item_head si_ih;
    char * si_dnm_data;
    struct si * si_next;

    // changed by XB;
    struct si * last_known;
};
void put_saved_items_into_tree (struct si *);
int reiserfsck_file_write (struct item_head * ih, char * item);
//int are_file_items_correct (struct key * key, int key_version, __u64 * size, int mark_passed_items, struct path *, void **);
int are_file_items_correct (struct key * key, int key_version, __u64 * size,  __u64 * min_size, __u32 * blocks, int mark_passed_items, int symlink);

/* pass2.c */
typedef	void (action_on_item_t)(struct si **, struct item_head *, char *);
action_on_item_t save_item;
action_on_item_t insert_item_separately;
void for_all_items_in_node (action_on_item_t action, struct si ** si, struct buffer_head * bh, int flag);
void take_bad_blocks_put_into_tree ();
void insert_each_item_separately (struct buffer_head *);


/* semantic.c */
extern struct key g_root_directory_key;
void semantic_pass (void);
int check_semantic_tree (struct key * key, struct key * parent, int is_dot_dot);



/* pass4.c */
int check_unaccessed_items (void);
void pass4 (struct super_block *);


/* check.c */
int bad_pair (struct super_block * s, struct buffer_head * bh, int i);
int bad_block_head (struct buffer_head * bh);
int check_file_system (void);
void reiserfsck_check_pass1 (void);
void reiserfsck_check_after_all (void);
int is_leaf_bad (struct buffer_head * bh);
int is_internal_bad (struct buffer_head * bh);

void check_fs_tree (struct super_block * s);


/* ustree.c */
void init_tb_struct (struct tree_balance * tb, struct super_block  * s, struct path * path, int size);
void reiserfsck_paste_into_item (struct path * path, const char * body, int size);
void reiserfsck_insert_item (struct path * path, struct item_head * ih, const char * body);
void reiserfsck_delete_item (struct path * path);
void reiserfsck_cut_from_item (struct path * path, int cut_size);
//typedef	int (comp_function_t)(void * key1, void * key2);
typedef	int (comp3_function_t)(void * key1, void * key2, int version);
typedef int (comp_function_t)(struct key * key1, struct key * key2);
int ubin_search_id(__u32 * id, __u32 * base, __u32 number, __u32 * pos);
int usearch_by_key (struct super_block * s, struct key * key, struct path * path);
int usearch_by_key_3 (struct super_block * s, struct key * key, struct path * path, int * repeat, int stop_level,
		      comp3_function_t comp_func, int version);		
int usearch_by_entry_key (struct super_block * s, struct key * key, struct path * path);
int usearch_by_position (struct super_block * s, struct key * key, int version, struct path * path);
struct key * uget_lkey (struct path * path);
struct key * uget_rkey (struct path * path);

typedef int do_after_read_t (struct super_block * s, struct buffer_head **, int h);
typedef int do_on_full_path_t (struct super_block * s, struct buffer_head **, int);
void pass_through_tree (struct super_block *, do_after_read_t, do_on_full_path_t);
//int comp_keys_3 (void * key1, void * key2);
//int comp_dir_entries (void * key1, void * key2);


/* bitmap.c */
extern int from_journal;
/*int reiserfs_new_blocknrs (	struct reiserfs_transaction_handle *th,
				struct super_block * s,
				unsigned long * free_blocknrs,
				unsigned long start,
				int amount_needed,
				int for_preserve_list); */
int reiserfs_new_blocknrs (	struct reiserfs_transaction_handle *th,
				unsigned long * pblocknrs,
				unsigned long start_from,
				int amount_needed);
//void reiserfs_free_block (struct reiserfs_transaction_handle *th, struct super_block * s, unsigned long block);
void reiserfs_free_block (struct reiserfs_transaction_handle *th, unsigned long);


void reiserfs_free_internal_block (struct super_block * s, unsigned long block);
struct buffer_head * reiserfsck_get_new_buffer (unsigned long start);
void force_freeing (void);
int is_block_used (unsigned long block);
int was_block_used (unsigned long block);
void mark_block_used (unsigned long block);
void mark_block_uninsertable (unsigned long block);
int is_block_uninsertable (unsigned long block);
void mark_block_unformatted (unsigned long block);
void mark_block_formatted (unsigned long block);
void unmark_block_unformatted (unsigned long block);
void unmark_block_formatted (unsigned long block);
int did_we_meet_it (struct super_block * s, blocknr_t block);

/* objectid.c */
int is_objectid_used (struct super_block * s, __u32 objectid);
void mark_objectid_as_used (struct super_block * s, __u32 objectid);
void mark_objectid_as_free (struct super_block * s, __u32 objectid);
__u32 get_unused_objectid (struct super_block * s);



/* segments.c */
struct overwritten_unfm_segment {
  int ous_begin;
  int ous_end;
  struct overwritten_unfm_segment * ous_next;  
};
struct overwritten_unfm * look_for_overwritten_unfm (__u32);
struct overwritten_unfm_segment * find_overwritten_unfm (unsigned long unfm, int length, struct overwritten_unfm_segment * segment_to_init);
int get_unoverwritten_segment (struct overwritten_unfm_segment * list_head, struct overwritten_unfm_segment * unoverwritten_segment);
void save_unfm_overwriting (unsigned long unfm, struct item_head * direct_ih);
void free_overwritten_unfms (void);
void mark_formatted_pointed_by_indirect (__u32);
int is_formatted_pointed_by_indirect (__u32);


/* do_balan.c */
/* lbalance.c */
/* ibalance.c */	/* links to fs/reiser */
/* fix_node.c */
/* teahash3.c */


/* info.c */
struct fsck_stat {
  /* pass 1,2 */
  int fs_good_leaves;
  int fs_uninsertable_leaves;
  int fs_rewritten_files;
  int fs_leaves_used_by_indirect_items;
  int fs_unfm_overwriting_unfm;
  int fs_indirect_to_direct;
  /* pass 3 */
  int fs_incorrect_regular_files;
  int fs_fixed_size_directories;
  int fs_fixed_size_files;
  int fs_deleted_entries;
  /* pass 4 */
  int fs_unaccessed_items;
  int fs_fixed_right_delim_key;
  /* fs stat */
  int fs_stat_data_items;
  int fs_regular_files;
  int fs_directories;
  int fs_symlinks;
  int fs_others;
};
  

extern struct fsck_stat g_fsck_info;

/* pass 1,2 */
#define GOOD_LEAVES 0
#define UNINSERTABLE_LEAVES 1
#define REWRITTEN_FILES 2
#define LEAVES_USED_BY_INDIRECT_ITEMS 3
#define UNFM_OVERWRITING_UNFM 4		/* overwrite contents of unformatted node keeping what has been written there from direct items */

/* pass 3 (semantic) */
#define INCORRECT_REGULAR_FILES 5
#define FIXED_SIZE_DIRECTORIES 6
#define FIXED_SIZE_FILES 7
#define DELETED_ENTRIES 8
#define INDIRECT_TO_DIRECT 9

extern inline int change_version (int version)
{
   return (version == 1)?0:1;
}

/* pass 4 */
#define UNACCESSED_ITEMS 10
#define FIXED_RIGHT_DELIM_KEY 11

/* fs stat */
#define STAT_DATA_ITEMS 12
#define REGULAR_FILES 13
#define SYMLINKS 14
#define OTHERS 15
#define DIRECTORIES 16

void add_event (int event);
int get_event (int event);
void output_information ();


/* journal.c */
void replay_all (struct super_block * s);

do_after_read_t release_journal_blocks;
void reset_journal (struct super_block * s);



struct what_to_do {
    int mode;			/* check, rebuild-tree, rebuild-sb */
    int do_journal_replay;	/* do journal replay or not */
    int verbose;       /* two levels of this only so far: verbose or not */
    int force; /* if --force specified with --rebuild-tree - will cause
		  rebuild even if filesystem seems clean */
    int yes_confirmed; /* does not work for rebuild-tree */
    int fix_fixable; /* fix things which can be fixed w/o rebuid - bitmap
			correction, for instance */
#if 0
    int where_to_stop;		/* not available */
    int what_to_scan;		/* not available */
#endif
};

struct id_map {
    __u32 * m_begin; // pointer to map area
    __u32   m_used;  // used places in area <= m_count * MAP_SIZE / sizeof (__u32)
    __u32   m_count; // count of MAPs
};			


/* fsck attaches this structure to filesystem handle's s_vp */
struct fsck_union {
    struct what_to_do * fu_job; /* options fsck is called with */

    reiserfs_bitmap_t * fu_check_bitmap; /* this bitmap is used running --check
                                           to notice blocks referred to more
                                           than once */
    unsigned int fu_used_blocks;

    int fu_fsck_stage;
    reiserfs_bitmap_t * fu_disk_bitmap; /* bitmaps needed by --rebuild-tree */
    reiserfs_bitmap_t * fu_new_bitmap;
    reiserfs_bitmap_t * fu_uninsertable_leaves; /* leaves pointed by
                                                   unformatted nodes */
    reiserfs_bitmap_t * fu_unformatted_bitmap;
    reiserfs_bitmap_t * fu_formatted_bitmap;


    struct id_map fu_id_map; /* map of used objectids built on semantic check */

    
};


/* the above structure is attached to reiserfs_filsys_t. Its fields are
   accessed by these */
#define fsck_union(fs) ((struct fsck_union *)((fs)->s_vp))
#define check_bitmap(fs) (fsck_union (fs)->fu_check_bitmap)
#define used_blocks(fs) (fsck_union (fs)->fu_used_blocks)
#define proper_id_map(fs) (fsck_union (fs)->fu_id_map)

/* on-disk bitmap */
#define disk_bitmap(fs) (fsck_union (fs)->fu_disk_bitmap)

/* bitmap which is being created during rebuild */
#define new_bitmap(fs) (fsck_union (fs)->fu_new_bitmap)

#define uninsertable_leaves(fs) (fsck_union (fs)->fu_uninsertable_leaves)
#define not_for_insertion(fs) uninsertable_leaves(fs)

/* after pass0 all unformatted node pointers are marked here */
#define unformatted_bitmap(fs) (fsck_union (fs)->fu_unformatted_bitmap)
#define found_unfm_ptrs(fs) unformatted_bitmap(fs)

/* after pass0 all nodes looking like leaves are marked here */
#define formatted_bitmap(fs) (fsck_union (fs)->fu_formatted_bitmap)
#define found_leaves(fs) (fsck_union (fs)->fu_formatted_bitmap)


#define fsck_stage(fs) (fsck_union (fs)->fu_fsck_stage)
