/*
 *  Copyright 2000 by Hans Reiser, licensing governed by reiserfs/README
 */

#ifndef REISERFS_LIB_H
#define REISERFS_LIB_H

#include <reiserfs_fs.h>

struct _bitmap {
    unsigned long bm_size;
    char * bm_map;
};

typedef struct _bitmap reiserfs_bitmap_t;


struct super_block {
    int s_version;		                  /* on-disk format version */
    reiserfs_bitmap_t * bitmap;			  /* copy of reiserfs on-disk bitmap */
    
    int s_dev; /* descriptor of opened block device file */
    int s_blocksize;
    struct buffer_head ** s_ap_bitmap; /* array of buffers containing bitmap
                                          blocks */
    struct buffer_head * s_sbh;  /* buffer containing super block */
    struct reiserfs_super_block * s_rs; /* pointer to its b_data */
    int s_dirt;
    hashf_t s_hash_function; /* pointer to function which is used to sort
				names in directory. It is set by reiserfs_open
				if it is set in the super block, otherwise it
				is set by first is_properly_hashed */
    int s_flags;
    void * s_vp;
};

typedef struct super_block * reiserfs_filsys_t;

/* reiserfslib.c */

reiserfs_filsys_t reiserfs_open (char * filename, int flags, int * error, void * vp);
void reiserfs_flush (reiserfs_filsys_t fs);
void reiserfs_free (reiserfs_filsys_t fs);
void reiserfs_close (reiserfs_filsys_t fs);

/* bitmap.c */

reiserfs_bitmap_t * reiserfs_create_bitmap (unsigned int bit_count);
int reiserfs_expand_bitmap (reiserfs_bitmap_t * bm, unsigned int bit_count);
void reiserfs_delete_bitmap (reiserfs_bitmap_t * bm);
void reiserfs_bitmap_set_bit (reiserfs_bitmap_t * bm, unsigned int bit_number);
void reiserfs_bitmap_clear_bit (reiserfs_bitmap_t * bm, unsigned int bit_number);

int reiserfs_bitmap_test_bit (reiserfs_bitmap_t * bm, unsigned int bit_number);
int reiserfs_bitmap_find_zero_bit (reiserfs_bitmap_t * bm, unsigned int * start);
int reiserfs_fetch_bitmap_from_disk (reiserfs_bitmap_t * bm, reiserfs_filsys_t * fs);
void reiserfs_bitmap_zero (reiserfs_bitmap_t * bm);
void reiserfs_bitmap_fill (reiserfs_bitmap_t * bm);

/* node_formats.c */

typedef int (*check_unfm_func_t) (reiserfs_filsys_t fs, __u32);
int is_it_bad_item (reiserfs_filsys_t, struct item_head *, char *,
		    check_unfm_func_t, int bad_dir);

#define hash_func_is_unknown(fs) ((fs)->s_hash_function == 0)
#define reiserfs_hash(fs) ((fs)->s_hash_function)
int hash_code (reiserfs_filsys_t);

#endif /* REISERFS_LIB_H */
